# This file was auto-generated by Fern from our API Definition.

import typing
from json.decoder import JSONDecodeError

from ..core.api_error import ApiError
from ..core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ..core.http_response import AsyncHttpResponse, HttpResponse
from ..core.jsonable_encoder import jsonable_encoder
from ..core.pydantic_utilities import parse_obj_as
from ..core.request_options import RequestOptions
from ..core.serialization import convert_and_respect_annotation_metadata
from ..errors.bad_request_error import BadRequestError
from ..errors.conflict_error import ConflictError
from ..errors.not_found_error import NotFoundError
from ..errors.unauthorized_error import UnauthorizedError
from ..types.agent_blueprint_public import AgentBlueprintPublic
from ..types.agent_blueprint_write import AgentBlueprintWrite
from ..types.agent_config_env import AgentConfigEnv
from ..types.blueprint_page_history import BlueprintPageHistory

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class RawAgentConfigsClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._client_wrapper = client_wrapper

    def create_agent_config(
        self,
        *,
        blueprint: AgentBlueprintWrite,
        project_id: typing.Optional[str] = OMIT,
        project_name: typing.Optional[str] = OMIT,
        id: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[None]:
        """
        Creates a new optimizer config with initial blueprint. Fails if the project already has a config.

        Parameters
        ----------
        blueprint : AgentBlueprintWrite

        project_id : typing.Optional[str]
            Project ID. Either project_id or project_name must be provided

        project_name : typing.Optional[str]
            Project name. Either project_id or project_name must be provided

        id : typing.Optional[str]
            Agent config ID. Generated automatically if not provided

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/private/agent-configs/blueprints",
            method="POST",
            json={
                "project_id": project_id,
                "project_name": project_name,
                "id": id,
                "blueprint": convert_and_respect_annotation_metadata(
                    object_=blueprint, annotation=AgentBlueprintWrite, direction="write"
                ),
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 409:
                raise ConflictError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def update_agent_config(
        self,
        *,
        blueprint: AgentBlueprintWrite,
        project_id: typing.Optional[str] = OMIT,
        project_name: typing.Optional[str] = OMIT,
        id: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[None]:
        """
        Adds a new blueprint to an existing optimizer config. Fails if the project has no config yet.

        Parameters
        ----------
        blueprint : AgentBlueprintWrite

        project_id : typing.Optional[str]
            Project ID. Either project_id or project_name must be provided

        project_name : typing.Optional[str]
            Project name. Either project_id or project_name must be provided

        id : typing.Optional[str]
            Agent config ID. Generated automatically if not provided

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/private/agent-configs/blueprints",
            method="PATCH",
            json={
                "project_id": project_id,
                "project_name": project_name,
                "id": id,
                "blueprint": convert_and_respect_annotation_metadata(
                    object_=blueprint, annotation=AgentBlueprintWrite, direction="write"
                ),
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def create_or_update_envs(
        self,
        *,
        project_id: str,
        envs: typing.Sequence[AgentConfigEnv],
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[None]:
        """
        Creates or updates environment-to-blueprint mappings

        Parameters
        ----------
        project_id : str

        envs : typing.Sequence[AgentConfigEnv]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/private/agent-configs/blueprints/environments",
            method="POST",
            json={
                "project_id": project_id,
                "envs": convert_and_respect_annotation_metadata(
                    object_=envs, annotation=typing.Sequence[AgentConfigEnv], direction="write"
                ),
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def get_blueprint_by_env(
        self,
        env_name: str,
        project_id: str,
        *,
        mask_id: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[AgentBlueprintPublic]:
        """
        Retrieves the blueprint associated with a specific environment

        Parameters
        ----------
        env_name : str

        project_id : str

        mask_id : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[AgentBlueprintPublic]
            Blueprint retrieved
        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/private/agent-configs/blueprints/environments/{jsonable_encoder(env_name)}/projects/{jsonable_encoder(project_id)}",
            method="GET",
            params={
                "mask_id": mask_id,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    AgentBlueprintPublic,
                    parse_obj_as(
                        type_=AgentBlueprintPublic,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def set_env_by_blueprint_name(
        self,
        env_name: str,
        project_id: str,
        *,
        blueprint_name: str,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[None]:
        """
        Sets an environment to point to a blueprint identified by name

        Parameters
        ----------
        env_name : str

        project_id : str

        blueprint_name : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/private/agent-configs/blueprints/environments/{jsonable_encoder(env_name)}/projects/{jsonable_encoder(project_id)}",
            method="PUT",
            json={
                "blueprint_name": blueprint_name,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def delete_env(
        self, env_name: str, project_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[None]:
        """
        Soft-deletes an environment by setting its ended_at timestamp

        Parameters
        ----------
        env_name : str

        project_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/private/agent-configs/blueprints/environments/{jsonable_encoder(env_name)}/projects/{jsonable_encoder(project_id)}",
            method="DELETE",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def get_blueprint_by_id(
        self,
        blueprint_id: str,
        *,
        mask_id: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[AgentBlueprintPublic]:
        """
        Retrieves a specific blueprint by its ID

        Parameters
        ----------
        blueprint_id : str

        mask_id : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[AgentBlueprintPublic]
            Blueprint retrieved
        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/private/agent-configs/blueprints/{jsonable_encoder(blueprint_id)}",
            method="GET",
            params={
                "mask_id": mask_id,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    AgentBlueprintPublic,
                    parse_obj_as(
                        type_=AgentBlueprintPublic,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def get_blueprint_by_name(
        self,
        project_id: str,
        name: str,
        *,
        mask_id: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[AgentBlueprintPublic]:
        """
        Retrieves a specific blueprint by its name within a project

        Parameters
        ----------
        project_id : str

        name : str

        mask_id : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[AgentBlueprintPublic]
            Blueprint retrieved
        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/private/agent-configs/blueprints/projects/{jsonable_encoder(project_id)}/names/{jsonable_encoder(name)}",
            method="GET",
            params={
                "mask_id": mask_id,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    AgentBlueprintPublic,
                    parse_obj_as(
                        type_=AgentBlueprintPublic,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def get_blueprint_history(
        self,
        project_id: str,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[BlueprintPageHistory]:
        """
        Retrieves paginated blueprint history for a project

        Parameters
        ----------
        project_id : str

        page : typing.Optional[int]

        size : typing.Optional[int]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[BlueprintPageHistory]
            History retrieved
        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/private/agent-configs/blueprints/history/projects/{jsonable_encoder(project_id)}",
            method="GET",
            params={
                "page": page,
                "size": size,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    BlueprintPageHistory,
                    parse_obj_as(
                        type_=BlueprintPageHistory,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def get_delta_by_id(
        self, blueprint_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[AgentBlueprintPublic]:
        """
        Retrieves only the changes (delta) introduced in a specific blueprint

        Parameters
        ----------
        blueprint_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[AgentBlueprintPublic]
            Delta retrieved
        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/private/agent-configs/blueprints/{jsonable_encoder(blueprint_id)}/deltas",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    AgentBlueprintPublic,
                    parse_obj_as(
                        type_=AgentBlueprintPublic,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def get_latest_blueprint(
        self,
        project_id: str,
        *,
        mask_id: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[AgentBlueprintPublic]:
        """
        Retrieves the latest blueprint for a project

        Parameters
        ----------
        project_id : str

        mask_id : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[AgentBlueprintPublic]
            Blueprint retrieved
        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/private/agent-configs/blueprints/latest/projects/{jsonable_encoder(project_id)}",
            method="GET",
            params={
                "mask_id": mask_id,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    AgentBlueprintPublic,
                    parse_obj_as(
                        type_=AgentBlueprintPublic,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)


class AsyncRawAgentConfigsClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._client_wrapper = client_wrapper

    async def create_agent_config(
        self,
        *,
        blueprint: AgentBlueprintWrite,
        project_id: typing.Optional[str] = OMIT,
        project_name: typing.Optional[str] = OMIT,
        id: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[None]:
        """
        Creates a new optimizer config with initial blueprint. Fails if the project already has a config.

        Parameters
        ----------
        blueprint : AgentBlueprintWrite

        project_id : typing.Optional[str]
            Project ID. Either project_id or project_name must be provided

        project_name : typing.Optional[str]
            Project name. Either project_id or project_name must be provided

        id : typing.Optional[str]
            Agent config ID. Generated automatically if not provided

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/private/agent-configs/blueprints",
            method="POST",
            json={
                "project_id": project_id,
                "project_name": project_name,
                "id": id,
                "blueprint": convert_and_respect_annotation_metadata(
                    object_=blueprint, annotation=AgentBlueprintWrite, direction="write"
                ),
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 409:
                raise ConflictError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def update_agent_config(
        self,
        *,
        blueprint: AgentBlueprintWrite,
        project_id: typing.Optional[str] = OMIT,
        project_name: typing.Optional[str] = OMIT,
        id: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[None]:
        """
        Adds a new blueprint to an existing optimizer config. Fails if the project has no config yet.

        Parameters
        ----------
        blueprint : AgentBlueprintWrite

        project_id : typing.Optional[str]
            Project ID. Either project_id or project_name must be provided

        project_name : typing.Optional[str]
            Project name. Either project_id or project_name must be provided

        id : typing.Optional[str]
            Agent config ID. Generated automatically if not provided

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/private/agent-configs/blueprints",
            method="PATCH",
            json={
                "project_id": project_id,
                "project_name": project_name,
                "id": id,
                "blueprint": convert_and_respect_annotation_metadata(
                    object_=blueprint, annotation=AgentBlueprintWrite, direction="write"
                ),
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def create_or_update_envs(
        self,
        *,
        project_id: str,
        envs: typing.Sequence[AgentConfigEnv],
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[None]:
        """
        Creates or updates environment-to-blueprint mappings

        Parameters
        ----------
        project_id : str

        envs : typing.Sequence[AgentConfigEnv]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/private/agent-configs/blueprints/environments",
            method="POST",
            json={
                "project_id": project_id,
                "envs": convert_and_respect_annotation_metadata(
                    object_=envs, annotation=typing.Sequence[AgentConfigEnv], direction="write"
                ),
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def get_blueprint_by_env(
        self,
        env_name: str,
        project_id: str,
        *,
        mask_id: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[AgentBlueprintPublic]:
        """
        Retrieves the blueprint associated with a specific environment

        Parameters
        ----------
        env_name : str

        project_id : str

        mask_id : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[AgentBlueprintPublic]
            Blueprint retrieved
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/private/agent-configs/blueprints/environments/{jsonable_encoder(env_name)}/projects/{jsonable_encoder(project_id)}",
            method="GET",
            params={
                "mask_id": mask_id,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    AgentBlueprintPublic,
                    parse_obj_as(
                        type_=AgentBlueprintPublic,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def set_env_by_blueprint_name(
        self,
        env_name: str,
        project_id: str,
        *,
        blueprint_name: str,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[None]:
        """
        Sets an environment to point to a blueprint identified by name

        Parameters
        ----------
        env_name : str

        project_id : str

        blueprint_name : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/private/agent-configs/blueprints/environments/{jsonable_encoder(env_name)}/projects/{jsonable_encoder(project_id)}",
            method="PUT",
            json={
                "blueprint_name": blueprint_name,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def delete_env(
        self, env_name: str, project_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[None]:
        """
        Soft-deletes an environment by setting its ended_at timestamp

        Parameters
        ----------
        env_name : str

        project_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/private/agent-configs/blueprints/environments/{jsonable_encoder(env_name)}/projects/{jsonable_encoder(project_id)}",
            method="DELETE",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def get_blueprint_by_id(
        self,
        blueprint_id: str,
        *,
        mask_id: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[AgentBlueprintPublic]:
        """
        Retrieves a specific blueprint by its ID

        Parameters
        ----------
        blueprint_id : str

        mask_id : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[AgentBlueprintPublic]
            Blueprint retrieved
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/private/agent-configs/blueprints/{jsonable_encoder(blueprint_id)}",
            method="GET",
            params={
                "mask_id": mask_id,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    AgentBlueprintPublic,
                    parse_obj_as(
                        type_=AgentBlueprintPublic,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def get_blueprint_by_name(
        self,
        project_id: str,
        name: str,
        *,
        mask_id: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[AgentBlueprintPublic]:
        """
        Retrieves a specific blueprint by its name within a project

        Parameters
        ----------
        project_id : str

        name : str

        mask_id : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[AgentBlueprintPublic]
            Blueprint retrieved
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/private/agent-configs/blueprints/projects/{jsonable_encoder(project_id)}/names/{jsonable_encoder(name)}",
            method="GET",
            params={
                "mask_id": mask_id,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    AgentBlueprintPublic,
                    parse_obj_as(
                        type_=AgentBlueprintPublic,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def get_blueprint_history(
        self,
        project_id: str,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[BlueprintPageHistory]:
        """
        Retrieves paginated blueprint history for a project

        Parameters
        ----------
        project_id : str

        page : typing.Optional[int]

        size : typing.Optional[int]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[BlueprintPageHistory]
            History retrieved
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/private/agent-configs/blueprints/history/projects/{jsonable_encoder(project_id)}",
            method="GET",
            params={
                "page": page,
                "size": size,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    BlueprintPageHistory,
                    parse_obj_as(
                        type_=BlueprintPageHistory,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def get_delta_by_id(
        self, blueprint_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[AgentBlueprintPublic]:
        """
        Retrieves only the changes (delta) introduced in a specific blueprint

        Parameters
        ----------
        blueprint_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[AgentBlueprintPublic]
            Delta retrieved
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/private/agent-configs/blueprints/{jsonable_encoder(blueprint_id)}/deltas",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    AgentBlueprintPublic,
                    parse_obj_as(
                        type_=AgentBlueprintPublic,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def get_latest_blueprint(
        self,
        project_id: str,
        *,
        mask_id: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[AgentBlueprintPublic]:
        """
        Retrieves the latest blueprint for a project

        Parameters
        ----------
        project_id : str

        mask_id : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[AgentBlueprintPublic]
            Blueprint retrieved
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/private/agent-configs/blueprints/latest/projects/{jsonable_encoder(project_id)}",
            method="GET",
            params={
                "mask_id": mask_id,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    AgentBlueprintPublic,
                    parse_obj_as(
                        type_=AgentBlueprintPublic,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)



import mimetypes
import random
import time
from typing import Optional


# The attachment file name regex
ATTACHMENT_FILE_NAME_REGEX = r"(?:input|output|metadata)-attachment-\d+-\d+-sdk\.\w+"
ATTACHMENT_FILE_NAME_PLACEHOLDER_REGEX = (
    r"\[((?:input|output|metadata)-attachment-\d+-\d+-sdk\.\w+)\]"
)


def get_file_extension(mime_type: str) -> str:
    """Convert MIME type to file extension.

    Mirrors the Java getFileExtension() method in AttachmentStripperService.

    Args:
        mime_type: The MIME type (e.g., "image/png", "application/pdf")

    Returns:
        File extension without a leading dot (e.g., "png", "pdf")
    """
    if not mime_type:
        return "bin"

    # Try to get extension from mimetypes module
    extension = mimetypes.guess_extension(mime_type, strict=False)

    if extension:
        # Remove the leading dot
        extension = extension.lstrip(".")
        # Handle special cases where mimetypes returns less common extensions
        if mime_type == "image/jpeg" and extension == "jpe":
            return "jpg"
        return extension

    # Fallback: extract from the MIME type (e.g., "image/png" -> "png")
    if "/" in mime_type:
        subtype = mime_type.split("/")[1]
        # Handle special cases like "svg+xml" -> "svg"
        if "+" in subtype:
            subtype = subtype.split("+")[0]
        # Remove any parameters (e.g., "jpeg; charset=utf-8" -> "jpeg")
        subtype = subtype.split(";")[0].strip()
        return subtype

    return "bin"


def detect_mime_type(data: bytes) -> Optional[str]:
    """Detect MIME type from byte content using magic bytes.

    This provides basic MIME type detection similar to Apache Tika in the Java implementation.
    It checks common file format magic bytes.

    Args:
        data: The byte data to analyze

    Returns:
        Detected MIME type string, or "application/octet-stream" if unknown
    """
    if len(data) < 4:
        return "application/octet-stream"

    # Check common file format magic bytes
    # PNG
    if data[:8] == b"\x89PNG\r\n\x1a\n":
        return "image/png"

    # JPEG
    if data[:2] == b"\xff\xd8" and data[-2:] == b"\xff\xd9":
        return "image/jpeg"

    # GIF
    if data[:6] in (b"GIF87a", b"GIF89a"):
        return "image/gif"

    # PDF
    if data[:4] == b"%PDF":
        return "application/pdf"

    # WebP
    if data[:4] == b"RIFF" and data[8:12] == b"WEBP":
        return "image/webp"

    # SVG (XML-based, check for SVG tag)
    try:
        text = data[:1024].decode("utf-8", errors="ignore")
        if "<svg" in text.lower():
            return "image/svg+xml"
    except Exception:
        pass

    # MP4
    if len(data) >= 12 and data[4:8] == b"ftyp":
        return "video/mp4"

    # JSON
    try:
        text = data[:100].decode("utf-8", errors="strict").strip()
        if text.startswith("{") or text.startswith("["):
            return "application/json"
    except Exception:
        pass

    # Default to octet-stream for unknown types
    return "application/octet-stream"


def create_attachment_filename(context: str, extension: str) -> str:
    """
    Generates a unique attachment filename based on the provided context and file extension.

    This function creates a filename by combining the given context, a randomly generated
    prefix to ensure uniqueness, the current timestamp in milliseconds, and the provided
    file extension. The generated filename aligns with the backend convention for naming
    attachments, which includes specific formatting and structure.

    Args:
        context: The context to use as the base for the filename (e.g., "input",
            "output", or "metadata").
        extension: The file extension to use for the filename (e.g., "png",
            "jpg", "txt").

    Returns:
        A generated filename string in the format
            "{context}-attachment-{random_prefix}-{timestamp}.{extension}".
    """
    # The backend has the following naming convention: r"\\[((?:input|output|metadata)-attachment-\\d+-\\d+\\.\\w+)\\]"
    # Example: [input-attachment-1-1704067200000.png]

    timestamp = int(round(time.time() * 1000))
    # we need to generate a large enough random prefix to avoid collisions
    random_prefix = random.randint(1, 99999999)
    return f"{context}-attachment-{random_prefix}-{timestamp}-sdk.{extension}"



"""
Configuration models for LLMJudge evaluator.

These Pydantic models represent the JSON structure used by the Opik backend
for online evaluations. The configuration is stored in the
`automation_rule_evaluators.code` column and used to execute LLM-as-a-judge
evaluations on the server side.

The schema aligns with the Java backend's LlmAsJudgeCode structure:
- model: LlmAsJudgeModelParameters (name, temperature, seed, custom_parameters)
- messages: List[LlmAsJudgeMessage] (role, content)
- variables: Map<String, String>
- schema: List[LlmAsJudgeOutputSchema] (name, type, description)
"""

from typing import Any, List, Literal, Optional

import pydantic


DEFAULT_MODEL_NAME = "gpt-5-nano"
DEFAULT_REASONING_EFFORT = "low"


class LLMJudgeModelConfig(pydantic.BaseModel):
    """
    Model configuration for LLMJudge.

    Matches backend's LlmAsJudgeModelParameters structure.
    """

    name: Optional[str] = None
    """The model name (e.g., 'gpt-4o', 'claude-3-opus'). Optional."""

    temperature: Optional[float] = None
    """Temperature for model generation."""

    seed: Optional[int] = None
    """Seed for reproducible generation."""

    custom_parameters: Optional[dict[str, Any]] = pydantic.Field(
        default=None, alias="customParameters"
    )
    """Optional custom parameters for the model."""

    model_config = pydantic.ConfigDict(populate_by_name=True)


class LLMJudgeMessage(pydantic.BaseModel):
    """A message in the LLMJudge prompt template.

    Matches backend's LlmAsJudgeMessage structure.
    """

    role: Literal["USER", "SYSTEM", "ASSISTANT"]
    """The role of the message sender."""

    content: str
    """The content of the message."""


class LLMJudgeSchemaItem(pydantic.BaseModel):
    """Schema definition for an assertion output.

    Matches backend's LlmAsJudgeOutputSchema(name, type, description).
    """

    name: str
    """The name of the assertion/score."""

    type: Literal["INTEGER", "DOUBLE", "BOOLEAN"]
    """The type of the score value (matches backend's LlmAsJudgeOutputSchemaType)."""

    description: str
    """Description of the expected behavior to check."""


CONFIG_VERSION = "1"


class LLMJudgeConfig(pydantic.BaseModel):
    """
    Configuration for LLMJudge evaluator.

    This structure represents the JSON format used by the Opik backend
    for online evaluations stored in `automation_rule_evaluators.code`.

    Note: The evaluator 'name' is stored separately in the automation_rule_evaluators
    table, not inside the code JSON. It's included here for SDK convenience.
    """

    version: str = CONFIG_VERSION
    """Schema version for forward-compatible deserialization."""

    name: str = "llm_judge"
    """The name of the evaluator (SDK convenience, not part of backend code JSON)."""

    model: LLMJudgeModelConfig
    """Model configuration with name, temperature, seed."""

    messages: List[LLMJudgeMessage]
    """Prompt template messages."""

    variables: dict[str, str]
    """Variable mappings from trace fields to template variables."""

    schema_: List[LLMJudgeSchemaItem] = pydantic.Field(alias="schema")
    """Output schema definitions for each assertion."""

    model_config = pydantic.ConfigDict(populate_by_name=True)



import contextlib
import logging
import os
from typing import Any, Dict, Iterator, List, Optional, Union

from haystack import tracing

from opik import tracing_runtime_config, url_helpers
from opik.decorator import arguments_helpers, span_creation_handler
from opik.api_objects import opik_client
from opik.api_objects import span as opik_span
from opik.api_objects import trace as opik_trace
from opik.types import SpanType

from . import opik_span_bridge
from . import constants

LOGGER = logging.getLogger(__name__)


class OpikTracer(tracing.Tracer):
    """Bridge between Haystack tracer and Opik for tracing pipeline operations."""

    def __init__(
        self,
        opik_client: opik_client.Opik,
        name: str = "Haystack",
        project_name: Optional[str] = None,
    ) -> None:
        """
        Initialize OpikTracer.

        Args:
            opik_client: The Opik client instance.
            name: The name of the pipeline or component.
            project_name: The name of the project for tracing (optional).
        """
        if not tracing.tracer.is_content_tracing_enabled:
            LOGGER.warning(
                "Traces will not be logged to Opik because Haystack tracing is disabled. "
                "To enable, set the HAYSTACK_CONTENT_TRACING_ENABLED environment variable to true "
                "before importing Haystack."
            )
        self._opik_client = opik_client
        self._context: List[opik_span_bridge.OpikSpanBridge] = []
        self._name = name
        self._project_name = project_name
        self.enforce_flush = (
            os.getenv(constants.HAYSTACK_OPIK_ENFORCE_FLUSH_ENV_VAR, "false").lower()
            == "true"
        )

    @contextlib.contextmanager
    def trace(
        self,
        operation_name: str,
        tags: Optional[Dict[str, Any]] = None,
        parent_span: Optional[opik_span_bridge.OpikSpanBridge] = None,
    ) -> Iterator[tracing.Span]:
        tags = tags or {}
        span_name = tags.get(constants.COMPONENT_NAME_KEY, operation_name)

        if parent_span:
            span = self._create_child_span(parent_span, span_name, operation_name, tags)
        else:
            span = self._create_span_or_trace(operation_name, span_name)

        self._context.append(span)

        if tags:
            span.set_tags(tags)

        try:
            yield span
        finally:
            self._finalize_span(span)

    def _create_span_or_trace(
        self, operation_name: str, span_name: str
    ) -> opik_span_bridge.OpikSpanBridge:
        """Create a span or trace based on existing context using span_creation_handler."""
        # For pipeline operations, use the pipeline name, otherwise use component name
        final_name = self._name if "pipeline.run" in operation_name else span_name
        metadata = {"created_from": "haystack", "operation": operation_name}

        # Always use span_creation_handler - it handles existing context properly
        start_span_parameters = arguments_helpers.StartSpanParameters(
            name=final_name,
            type="general",
            metadata=metadata,
            project_name=self._project_name,
        )

        result = span_creation_handler.create_span_respecting_context(
            start_span_arguments=start_span_parameters,
            distributed_trace_headers=None,
        )
        final_span_or_trace_data: Union[opik_span.SpanData, opik_trace.TraceData] = (
            result.trace_data if result.trace_data is not None else result.span_data
        )

        return opik_span_bridge.OpikSpanBridge(final_span_or_trace_data)

    def _create_child_span(
        self,
        parent_span: opik_span_bridge.OpikSpanBridge,
        span_name: str,
        operation_name: str,
        tags: Dict[str, Any],
    ) -> opik_span_bridge.OpikSpanBridge:
        """Create a child span from a parent span."""
        parent_data = parent_span.get_opik_span_or_trace_data()
        span_type: SpanType = (
            "llm"
            if tags.get(constants.COMPONENT_TYPE_KEY)
            in constants.ALL_SUPPORTED_GENERATORS
            else "general"
        )

        span_data = parent_data.create_child_span_data(
            name=span_name,
            type=span_type,
            metadata={"created_from": "haystack", "operation": operation_name},
        )
        return opik_span_bridge.OpikSpanBridge(span_data)

    def _finalize_span(self, span: opik_span_bridge.OpikSpanBridge) -> None:
        """Finalize span by updating metadata and ending the span."""
        try:
            span_or_trace_data = span.get_opik_span_or_trace_data()
            span.apply_component_metadata()
            span_or_trace_data.init_end_time()

            # Send data to backend if tracing is active
            if tracing_runtime_config.is_tracing_active():
                if isinstance(span_or_trace_data, opik_trace.TraceData):
                    self._opik_client.__internal_api__trace__(
                        **span_or_trace_data.as_parameters
                    )
                else:
                    self._opik_client.__internal_api__span__(
                        **span_or_trace_data.as_parameters
                    )

            self._context.pop()
            if self.enforce_flush:
                self.flush()
        except Exception as e:
            LOGGER.error("Failed to finalize span: %s", e, exc_info=True)
            if self._context:
                self._context.pop()

    def flush(self) -> None:
        """Flush the Opik client to send pending data."""
        self._opik_client.flush()

    def current_span(self) -> Optional[opik_span_bridge.OpikSpanBridge]:
        """Return the current active span."""
        return self._context[-1] if self._context else None

    def get_project_url(self) -> Optional[str]:
        """Return the URL to the tracing data."""
        span = self.current_span()
        if not span:
            return None

        span_data = span.get_opik_span_or_trace_data()
        trace_id = (
            span_data.id
            if isinstance(span_data, opik_trace.TraceData)
            else span_data.trace_id
        )
        return url_helpers.get_project_url_by_trace_id(
            trace_id=trace_id, url_override=self._opik_client.config.url_override
        )

    def get_trace_id(self) -> Optional[str]:
        """Return the trace id of the current trace."""
        span = self.current_span()
        if not span:
            return None

        span_data = span.get_opik_span_or_trace_data()
        return (
            span_data.id
            if isinstance(span_data, opik_trace.TraceData)
            else span_data.trace_id
        )



import logging
from typing import NamedTuple, Optional

import httpx

from opik import httpx_client, url_helpers


LOGGER = logging.getLogger(__name__)


class ProbeResult(NamedTuple):
    is_healthy: bool
    error_message: Optional[str]


class ConnectionProbe:
    """
    Provides functionality to probe and verify the connection to a remote server.

    This class is designed to check the availability and responsiveness of a server
    by performing a lightweight connection test to a specific endpoint. It encapsulates
    the logic to send a request and interpret its response, ensuring it is simple to
    evaluate the connection status programmatically.
    """

    def __init__(self, base_url: str, client: httpx_client.OpikHttpxClient):
        """
        Initializes an instance with a base URL and an HTTP client.

        This constructor sets up the necessary URLs and initializes the
        client needed to make requests.

        Args:
            base_url: The base URL for API requests.
            client: The HTTP client instance to make API calls.
        """
        self._ping_url = url_helpers.get_is_alive_ping_url(base_url)
        self._client = client

    def check_connection(self, timeout: float) -> ProbeResult:
        """
        Checks the connection to a specified server by sending a ping request and evaluates the response.

        This method performs a basic health check by sending an HTTP GET request to the server's ping URL.
        If the response status code is 200, it indicates the server is healthy, otherwise it is considered unhealthy.
        Handles timeouts and other unexpected exceptions during the connection attempt.

        Args:
            timeout: The maximum duration, in seconds, to wait for the server's response before timing out.

        Returns:
            An object indicating whether the server is healthy and, if not, the associated error message.
        """
        try:
            response = self._client.get(self._ping_url, timeout=timeout)
            if response.status_code == 200:
                return ProbeResult(is_healthy=True, error_message=None)
            else:
                return ProbeResult(
                    is_healthy=False,
                    error_message=f"Unexpected status code: {response.status_code}",
                )
        except (httpx.ConnectError, httpx.TimeoutException) as e:
            return ProbeResult(is_healthy=False, error_message=f"Connection error: {e}")
        except Exception as e:
            LOGGER.exception(
                "Unexpected error while checking connection to server: %s",
                e,
                exc_info=True,
            )
            return ProbeResult(is_healthy=False, error_message=f"Unexpected error: {e}")



import dataclasses
import json
import typing

from opik.api_objects.prompt.base_prompt import BasePrompt
from opik.rest_api.types.prompt_version_detail import PromptVersionDetail

SUPPORTED_PRIMITIVE_TYPES = (str, int, float, bool)

_PYTHON_TO_BACKEND_TYPE: typing.Dict[type, str] = {
    str: "string",
    int: "integer",
    float: "float",
    bool: "boolean",
}

_BACKEND_TO_PYTHON_TYPE: typing.Dict[str, type] = {
    v: k for k, v in _PYTHON_TO_BACKEND_TYPE.items()
}


def backend_type_to_python_type(backend_type: str) -> typing.Optional[type]:
    return _BACKEND_TO_PYTHON_TYPE.get(backend_type)


def unwrap_optional(py_type: typing.Any) -> typing.Optional[typing.Any]:
    """Return the inner type of Optional[T] (i.e. Union[T, None]), or None if not Optional."""
    if typing.get_origin(py_type) is typing.Union:
        args = [a for a in typing.get_args(py_type) if a is not type(None)]
        if len(args) == 1:
            return args[0]
    return None


def is_prompt_type(py_type: typing.Any) -> bool:
    return isinstance(py_type, type) and issubclass(py_type, BasePrompt)


def is_prompt_version_type(py_type: typing.Any) -> bool:
    return py_type is PromptVersionDetail


def is_supported_type(py_type: typing.Any) -> bool:
    inner = unwrap_optional(py_type)
    if inner is not None:
        return is_supported_type(inner)
    if py_type in SUPPORTED_PRIMITIVE_TYPES:
        return True
    if is_prompt_type(py_type):
        return True
    if is_prompt_version_type(py_type):
        return True
    origin = typing.get_origin(py_type)
    if origin is list:
        args = typing.get_args(py_type)
        return len(args) == 1 and args[0] in SUPPORTED_PRIMITIVE_TYPES
    if origin is dict:
        args = typing.get_args(py_type)
        return (
            len(args) == 2 and args[0] is str and args[1] in SUPPORTED_PRIMITIVE_TYPES
        )
    return False


def python_type_to_backend_type(py_type: typing.Any) -> str:
    inner = unwrap_optional(py_type)
    if inner is not None:
        return python_type_to_backend_type(inner)
    if py_type in _PYTHON_TO_BACKEND_TYPE:
        return _PYTHON_TO_BACKEND_TYPE[py_type]
    if is_prompt_type(py_type):
        return "prompt"
    if is_prompt_version_type(py_type):
        return "prompt_commit"
    origin = typing.get_origin(py_type)
    if origin in (list, dict):
        return "string"
    raise TypeError(f"Unsupported type: {py_type}")


def python_value_to_backend_value(
    value: typing.Any, py_type: typing.Any
) -> typing.Optional[str]:
    if value is None:
        return None
    if py_type is bool:
        return "true" if value else "false"
    if py_type in (int, float):
        return str(value)
    if py_type is str:
        return value
    if is_prompt_type(py_type):
        return value.commit
    if is_prompt_version_type(py_type):
        return value.commit
    origin = typing.get_origin(py_type)
    if origin in (list, dict):
        return json.dumps(value)
    raise TypeError(f"Unsupported type: {py_type}")


def python_value_to_metadata_value(
    value: typing.Any, py_type: typing.Any
) -> typing.Any:
    if is_prompt_type(py_type) or is_prompt_version_type(py_type):
        return python_value_to_backend_value(value, py_type)
    return value


def backend_value_to_python_value(
    value: typing.Any,
    backend_type: str,
    py_type: typing.Any,
) -> typing.Any:
    if value is None:
        return None

    if py_type is bool:
        if isinstance(value, bool):
            return value
        return str(value).lower() == "true"

    if py_type is int:
        return int(float(value)) if not isinstance(value, int) else value

    if py_type is float:
        return float(value) if not isinstance(value, float) else value

    if py_type is str:
        return str(value)

    if is_prompt_type(py_type):
        return str(value)

    if is_prompt_version_type(py_type):
        return str(value)

    origin = typing.get_origin(py_type)
    if origin in (list, dict):
        if isinstance(value, str):
            return json.loads(value)
        return value

    raise TypeError(f"Unsupported type: {py_type}")


def extract_dataclass_fields(
    cls: type,
) -> typing.List[typing.Tuple[str, typing.Any, typing.Optional[str]]]:
    """Returns (field_name, field_type, description) for supported fields."""
    if not dataclasses.is_dataclass(cls):
        raise TypeError(f"{cls} is not a dataclass")

    type_hints = typing.get_type_hints(cls, include_extras=True)
    result = []
    for f in dataclasses.fields(cls):
        raw_hint = type_hints.get(f.name, f.type)
        description: typing.Optional[str] = None
        if typing.get_origin(raw_hint) is typing.Annotated:
            args = typing.get_args(raw_hint)
            py_type = args[0]
            description = next((a for a in args[1:] if isinstance(a, str)), None)
        else:
            py_type = raw_hint
        inner = unwrap_optional(py_type)
        if inner is not None:
            py_type = inner
        if not is_supported_type(py_type):
            continue
        result.append((f.name, py_type, description))
    return result



from typing import Optional, Union, Dict, Any

import pydantic

from opik.types import LLMProvider

from . import opik_usage


class LLMUsageInfo(pydantic.BaseModel):
    provider: Optional[Union[LLMProvider, str]] = None
    model: Optional[str] = None
    """
    The model name (or model id) that can be used to identify the model for cost calculation.
    """
    usage: Optional[Union[Dict[str, Any], opik_usage.OpikUsage]] = None
    """
    The dictionary with token usage in the original format of the provider.
    """



import os
import platform
import uuid
from typing import Optional, Tuple

import click
import httpx

from opik import Opik
from opik.rest_api.core.api_error import ApiError


@click.command(context_settings={"ignore_unknown_options": True})
@click.option("--pair", "pair_code", default=None, help="Pairing code for the runner.")
@click.option("--name", default=None, help="Runner name.")
@click.argument("command", nargs=-1, type=click.UNPROCESSED, required=False)
@click.pass_context
def connect(
    ctx: click.Context,
    pair_code: Optional[str],
    name: Optional[str],
    command: Tuple[str, ...],
) -> None:
    """Connect a local runner to Opik and exec the user command."""
    api_key = ctx.obj.get("api_key") if ctx.obj else None

    client = Opik(api_key=api_key, _show_misconfiguration_message=False)
    api = client.rest_client

    try:
        runner_name = name or f"{platform.node()}-{uuid.uuid4().hex[:6]}"
        resp = api.runners.connect_runner(
            runner_name=runner_name,
            pairing_code=pair_code,
        )

        runner_id = resp.runner_id
        if not runner_id:
            click.echo("Error: server did not return a runner_id")
            raise SystemExit(1)

        project_name = resp.project_name
        if not project_name:
            click.echo("Error: server did not return a project_name")
            raise SystemExit(1)

        click.echo(f"Runner connected (ID: {runner_id}).")

        if not command:
            click.echo("No command specified. Set env vars and exiting.")
            return

        env = {
            **os.environ,
            "OPIK_RUNNER_MODE": "true",
            "OPIK_RUNNER_ID": runner_id,
            "OPIK_PROJECT_NAME": project_name,
        }

        client.end()
        os.execvpe(command[0], list(command), env)
    except ApiError as e:
        click.echo(f"Error: {e.body}" if e.body else f"Error: {e.status_code}")
        raise SystemExit(1)
    except httpx.ConnectError:
        config = client.config
        click.echo(
            f"Error: Could not connect to Opik at {config.url_override}. "
            "Check that the backend is running."
        )
        raise SystemExit(1)
    except OSError as e:
        click.echo(f"Error: Could not execute command '{command[0]}': {e}")
        raise SystemExit(1)
    finally:
        client.end()



# This file was auto-generated by Fern from our API Definition.

# isort: skip_file

from .types import ExperimentWriteEvaluationMethod, ExperimentWriteStatus, ExperimentWriteType

__all__ = ["ExperimentWriteEvaluationMethod", "ExperimentWriteStatus", "ExperimentWriteType"]



"""SimulatedUser class for multi-turn conversation simulation."""

from typing import List, Dict, Optional
from opik.evaluation.models.models_factory import (
    get as get_model,
    get_default_model_name,
)


class SimulatedUser:
    """
    A simulated user that generates responses using LLMs or fixed responses.

    The user simulator generates string responses that are then incorporated
    into the conversation by the application logic.
    """

    def __init__(
        self,
        persona: str,
        model: Optional[str] = None,
        fixed_responses: Optional[List[str]] = None,
    ):
        """
        Initialize a simulated user.

        Args:
            persona: Description of the user's personality and behavior
            model: LLM model to use for generating responses. If omitted, uses the
                central default model resolved by `get_default_model_name()`
                (override with `OPIK_DEFAULT_LLM`).
            fixed_responses: Optional list of predefined responses to cycle through
        """
        self.persona = persona
        self.model = model or get_default_model_name()
        self.fixed_responses = fixed_responses or []
        self._response_index = 0

        # Initialize LLM backend using models_factory for consistency
        self._llm = get_model(model_name=self.model)

    def generate_response(self, conversation_history: List[Dict[str, str]]) -> str:
        """
        Generate a response based on the conversation history.

        Args:
            conversation_history: List of message dicts with 'role' and 'content' keys

        Returns:
            String response from the simulated user
        """
        # Use fixed responses first if available
        if self.fixed_responses:
            response = self.fixed_responses[
                self._response_index % len(self.fixed_responses)
            ]
            self._response_index += 1
            return response

        # Generate response using LLM
        return self._generate_llm_response(conversation_history)

    def _generate_llm_response(self, conversation_history: List[Dict[str, str]]) -> str:
        """Generate response using the LLM backend."""
        # Build system prompt with persona and clear instructions
        system_prompt = f"""You are a simulated user with the following persona: {self.persona}

Your task is to generate realistic user messages that this persona would send in a conversation.
Respond as if you are the user, not as an assistant describing the user.
Generate a single user message that fits your persona and the conversation context."""

        # Convert conversation history to messages format expected by LLM
        messages = [{"role": "system", "content": system_prompt}]

        # Add all conversation history
        messages.extend(conversation_history)

        # Convert messages to string format for generate_string
        conversation_text = self._format_messages_as_text(messages)

        # Generate response
        try:
            response = self._llm.generate_string(input=conversation_text)
            return response
        except Exception as e:
            # Fallback response if LLM fails
            return f"I'm having trouble responding right now. ({str(e)})"

    def _format_messages_as_text(self, messages: List[Dict[str, str]]) -> str:
        """Convert message list to text format for LLM input."""
        formatted_messages = []
        for message in messages:
            role = message["role"]
            content = message["content"]
            if role == "system":
                formatted_messages.append(f"System: {content}")
            elif role == "user":
                formatted_messages.append(f"User: {content}")
            elif role == "assistant":
                formatted_messages.append(f"Assistant: {content}")
            else:
                formatted_messages.append(f"{role.title()}: {content}")

        return "\n".join(formatted_messages)



# This file was auto-generated by Fern from our API Definition.

import typing
from json.decoder import JSONDecodeError

from ..core.api_error import ApiError
from ..core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ..core.http_response import AsyncHttpResponse, HttpResponse
from ..core.pydantic_utilities import parse_obj_as
from ..core.request_options import RequestOptions
from ..errors.bad_gateway_error import BadGatewayError
from ..errors.internal_server_error import InternalServerError
from ..errors.service_unavailable_error import ServiceUnavailableError
from ..errors.unprocessable_entity_error import UnprocessableEntityError
from ..types.error_message import ErrorMessage
from ..types.ollama_connection_test_response import OllamaConnectionTestResponse
from ..types.ollama_model import OllamaModel

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class RawOllamaClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._client_wrapper = client_wrapper

    def list_models(
        self,
        *,
        base_url: str,
        api_key: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[typing.List[OllamaModel]]:
        """
        Fetches the list of models available from the Ollama instance. URL may be provided with or without /v1 suffix (e.g., http://localhost:11434 or http://localhost:11434/v1). The /v1 suffix will be automatically removed for model discovery. For actual LLM inference, use the URL with /v1 suffix for OpenAI-compatible endpoints.

        Parameters
        ----------
        base_url : str
            Base URL of the Ollama instance. May include /v1 suffix which will be automatically removed for connection testing. For inference, use the URL with /v1 suffix for OpenAI-compatible endpoints.

        api_key : typing.Optional[str]
            Optional API key for authenticated Ollama instances. If provided, will be sent as Bearer token in Authorization header.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[typing.List[OllamaModel]]
            Models retrieved successfully
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/private/ollama/models",
            method="POST",
            json={
                "base_url": base_url,
                "api_key": api_key,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    typing.List[OllamaModel],
                    parse_obj_as(
                        type_=typing.List[OllamaModel],  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 503:
                raise ServiceUnavailableError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        ErrorMessage,
                        parse_obj_as(
                            type_=ErrorMessage,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def test_connection(
        self,
        *,
        base_url: str,
        api_key: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[OllamaConnectionTestResponse]:
        """
        Validates that the provided Ollama URL is reachable. URL may be provided with or without /v1 suffix (e.g., http://localhost:11434 or http://localhost:11434/v1). The /v1 suffix will be automatically removed for connection testing. For inference, use the URL with /v1 suffix.

        Parameters
        ----------
        base_url : str
            Base URL of the Ollama instance. May include /v1 suffix which will be automatically removed for connection testing. For inference, use the URL with /v1 suffix for OpenAI-compatible endpoints.

        api_key : typing.Optional[str]
            Optional API key for authenticated Ollama instances. If provided, will be sent as Bearer token in Authorization header.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[OllamaConnectionTestResponse]
            Connection test successful
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/private/ollama/test-connection",
            method="POST",
            json={
                "base_url": base_url,
                "api_key": api_key,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    OllamaConnectionTestResponse,
                    parse_obj_as(
                        type_=OllamaConnectionTestResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 502:
                raise BadGatewayError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        OllamaConnectionTestResponse,
                        parse_obj_as(
                            type_=OllamaConnectionTestResponse,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 503:
                raise ServiceUnavailableError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        ErrorMessage,
                        parse_obj_as(
                            type_=ErrorMessage,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)


class AsyncRawOllamaClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._client_wrapper = client_wrapper

    async def list_models(
        self,
        *,
        base_url: str,
        api_key: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[typing.List[OllamaModel]]:
        """
        Fetches the list of models available from the Ollama instance. URL may be provided with or without /v1 suffix (e.g., http://localhost:11434 or http://localhost:11434/v1). The /v1 suffix will be automatically removed for model discovery. For actual LLM inference, use the URL with /v1 suffix for OpenAI-compatible endpoints.

        Parameters
        ----------
        base_url : str
            Base URL of the Ollama instance. May include /v1 suffix which will be automatically removed for connection testing. For inference, use the URL with /v1 suffix for OpenAI-compatible endpoints.

        api_key : typing.Optional[str]
            Optional API key for authenticated Ollama instances. If provided, will be sent as Bearer token in Authorization header.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[typing.List[OllamaModel]]
            Models retrieved successfully
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/private/ollama/models",
            method="POST",
            json={
                "base_url": base_url,
                "api_key": api_key,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    typing.List[OllamaModel],
                    parse_obj_as(
                        type_=typing.List[OllamaModel],  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 503:
                raise ServiceUnavailableError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        ErrorMessage,
                        parse_obj_as(
                            type_=ErrorMessage,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def test_connection(
        self,
        *,
        base_url: str,
        api_key: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[OllamaConnectionTestResponse]:
        """
        Validates that the provided Ollama URL is reachable. URL may be provided with or without /v1 suffix (e.g., http://localhost:11434 or http://localhost:11434/v1). The /v1 suffix will be automatically removed for connection testing. For inference, use the URL with /v1 suffix.

        Parameters
        ----------
        base_url : str
            Base URL of the Ollama instance. May include /v1 suffix which will be automatically removed for connection testing. For inference, use the URL with /v1 suffix for OpenAI-compatible endpoints.

        api_key : typing.Optional[str]
            Optional API key for authenticated Ollama instances. If provided, will be sent as Bearer token in Authorization header.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[OllamaConnectionTestResponse]
            Connection test successful
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/private/ollama/test-connection",
            method="POST",
            json={
                "base_url": base_url,
                "api_key": api_key,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    OllamaConnectionTestResponse,
                    parse_obj_as(
                        type_=OllamaConnectionTestResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 502:
                raise BadGatewayError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        OllamaConnectionTestResponse,
                        parse_obj_as(
                            type_=OllamaConnectionTestResponse,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 503:
                raise ServiceUnavailableError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        ErrorMessage,
                        parse_obj_as(
                            type_=ErrorMessage,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)



import logging
import opik._logging as _logging
from typing import List, Any, Generator, TYPE_CHECKING
from opik.types import BatchFeedbackScoreDict

from opik.api_objects import opik_client
from . import test_runs_storage, experiment_runner, summary

import pytest

if TYPE_CHECKING:
    import _pytest.terminal

LOGGER = logging.getLogger(__name__)


def _is_plugin_enabled(config: Any) -> bool:
    option_enabled = bool(getattr(getattr(config, "option", None), "opik", False))

    ini_enabled = False
    if config is not None and hasattr(config, "getini"):
        try:
            ini_enabled = bool(config.getini("opik_pytest_enabled"))
        except ValueError:
            ini_enabled = False

    active_by_collection = bool(getattr(config, "_opik_pytest_active", False))

    return option_enabled or ini_enabled or active_by_collection


def pytest_addoption(parser: "pytest.Parser") -> None:
    group = parser.getgroup("opik")
    group.addoption(
        "--opik",
        action="store_true",
        default=False,
        help="Enable Opik pytest plugin hooks.",
    )
    parser.addini(
        "opik_pytest_enabled",
        "Enable Opik pytest plugin hooks.",
        type="bool",
        default=False,
    )


def pytest_collection_modifyitems(
    session: "pytest.Session", config: "pytest.Config", items: List["pytest.Item"]
) -> None:
    if _is_plugin_enabled(config):
        return

    for item in items:
        test_obj = getattr(item, "obj", None)
        if bool(getattr(test_obj, "_opik_llm_unit", False)):
            setattr(config, "_opik_pytest_active", True)
            return


@pytest.hookimpl(hookwrapper=True)
def pytest_runtest_makereport(item: "pytest.Item") -> Generator:
    """
    Write the results of each test in the session
    """
    plugin_enabled = _is_plugin_enabled(item.config)
    outcome = yield

    if not plugin_enabled:
        return

    try:
        if outcome is None:
            return
        report = outcome.get_result()
        if report.when == "call":
            item.report = report
    except Exception:
        LOGGER.debug(
            "Unexpected failure during opik pytest_runtest_makereport hook",
            exc_info=True,
        )


@_logging.convert_exception_to_log_message(
    "Unexpected failure during opik pytest_sessionfinish hook",
    logger=LOGGER,
    exc_info=True,
    logging_level=logging.ERROR,
)
def pytest_sessionfinish(session: "pytest.Session", exitstatus: Any) -> None:
    if not _is_plugin_enabled(session.config):
        return

    session_items: List["pytest.Item"] = list(getattr(session, "items", []) or [])
    llm_test_items: List["pytest.Item"] = [
        test_item
        for test_item in session_items
        if getattr(test_item, "nodeid", None) in test_runs_storage.LLM_UNIT_TEST_RUNS
    ]
    if len(llm_test_items) == 0:
        return

    try:
        traces_feedback_scores: List[BatchFeedbackScoreDict] = []
        test_items_with_reports: List["pytest.Item"] = []

        for item in llm_test_items:
            nodeid = getattr(item, "nodeid", None)
            if nodeid is None:
                continue

            report: Any = getattr(item, "report", None)
            trace_data = test_runs_storage.TEST_RUNS_TO_TRACE_DATA.get(nodeid)
            if report is None or trace_data is None:
                continue

            traces_feedback_scores.append(
                BatchFeedbackScoreDict(
                    id=trace_data.id,
                    name="Passed",
                    value=float(getattr(report, "passed", False)),
                )
            )
            test_items_with_reports.append(item)

        if len(test_items_with_reports) == 0:
            return

        client = opik_client.get_client_cached()
        try:
            client.log_traces_feedback_scores(traces_feedback_scores)
            experiment_runner.run(client=client, test_items=test_items_with_reports)
        finally:
            client.flush()
    except Exception:
        LOGGER.error(
            "Unexpected exception occured while trying to log LLM unit tests experiment results",
            exc_info=True,
        )


@_logging.convert_exception_to_log_message(
    "Unexpected exception occured while trying to print LLM unit tests summary",
    logger=LOGGER,
    exc_info=True,
    logging_level=logging.DEBUG,
)
def pytest_terminal_summary(
    terminalreporter: "_pytest.terminal.TerminalReporter",
) -> None:
    if not _is_plugin_enabled(terminalreporter.config):
        return

    reports: List[pytest.TestReport] = terminalreporter.stats.get(
        "passed", []
    ) + terminalreporter.stats.get("failed", [])

    llm_reports = [
        report
        for report in reports
        if report.nodeid in test_runs_storage.LLM_UNIT_TEST_RUNS
    ]

    if len(llm_reports) == 0:
        return

    summary.print(llm_reports)



import threading
from typing import Type, Dict
from .. import messages
from . import base_batcher
from . import flushing_thread


class BatchManager:
    def __init__(
        self,
        message_to_batcher_mapping: Dict[
            Type[messages.BaseMessage], base_batcher.BaseBatcher
        ],
    ) -> None:
        self._message_to_batcher_mapping = message_to_batcher_mapping
        self._flushing_thread = flushing_thread.FlushingThread(
            batchers=list(self._message_to_batcher_mapping.values())
        )
        self._lock = threading.RLock()

    def start(self) -> None:
        self._flushing_thread.start()

    def stop(self) -> None:
        with self._lock:
            # stop the flushing thread
            self._flushing_thread.close()
            # force flush all pending messages
            self.flush()

    def message_supports_batching(self, message: messages.BaseMessage) -> bool:
        if message is None:
            return False

        if hasattr(message, "supports_batching"):
            return message.supports_batching

        return message.__class__ in self._message_to_batcher_mapping

    def process_message(self, message: messages.BaseMessage) -> None:
        with self._lock:
            self._message_to_batcher_mapping[type(message)].add(message)

    def is_empty(self) -> bool:
        with self._lock:
            return all(
                [
                    batcher.is_empty()
                    for batcher in self._message_to_batcher_mapping.values()
                ]
            )

    def flush(self) -> None:
        with self._lock:
            for batcher in self._message_to_batcher_mapping.values():
                batcher.flush()



from typing import Any, Optional

from .. import base_metric, score_result


class Contains(base_metric.BaseMetric):
    """
    A metric that checks if a reference string is contained within an output string.

    This metric returns a score of 1.0 if the reference string is found within the output string,
    and 0.0 otherwise. The comparison can be made case-sensitive or case-insensitive.

    Args:
        case_sensitive: Whether the comparison should be case-sensitive. Defaults to False.
        reference: Optional default reference string. If provided, it will be used unless
            a reference is explicitly passed to `score()`.
        name: The name of the metric. Defaults to "contains_metric".
        track: Whether to track the metric. Defaults to True.
        project_name: Optional project name to track the metric in for the cases when there are
            no parent span/trace to inherit project name from.

    Examples:
        >>> # Using a default reference at initialization
        >>> contains_metric = Contains(reference="world")
        >>> result = contains_metric.score("Hello, World!")
        >>> print(result.value)
        1.0

        >>> # Overriding the default reference at score time
        >>> result = contains_metric.score("Hello, World!", reference="there")
        >>> print(result.value)
        0.0

        >>> # If no reference is set at all, score() raises an error
        >>> contains_metric = Contains()
        >>> contains_metric.score("Hello")
        Traceback (most recent call last):
            ...
        ValueError: No reference string provided. Either pass `reference` to `score()` or set a default reference when creating the metric.

        >>> # Empty reference string is invalid
        >>> contains_metric = Contains(reference="")
        >>> contains_metric.score("Hello")
        Traceback (most recent call last):
            ...
        ValueError: Invalid reference string provided. Reference must be a non-empty string.
    """

    def __init__(
        self,
        case_sensitive: bool = False,
        reference: Optional[str] = None,
        name: str = "contains_metric",
        track: bool = True,
        project_name: Optional[str] = None,
    ):
        super().__init__(
            name=name,
            track=track,
            project_name=project_name,
        )
        self._case_sensitive = case_sensitive
        self._default_reference = reference

    def score(
        self, output: str, reference: Optional[str] = None, **ignored_kwargs: Any
    ) -> score_result.ScoreResult:
        """
        Calculate the score based on whether the reference string is contained in the output string.

        Args:
            output: The output string to check.
            reference: The reference string to look for in the output. If None, falls back to the
                default reference provided at initialization.
            **ignored_kwargs: Additional keyword arguments that are ignored.

        Returns:
            score_result.ScoreResult: A ScoreResult object with a value of 1.0 if the reference
                is found in the output, 0.0 otherwise.
        """
        # Use provided reference, else fall back to default
        ref = reference if reference is not None else self._default_reference

        # Handle missing reference (None) separately
        if ref is None:
            raise ValueError(
                "No reference string provided. Either pass `reference` to `score()` or set a default reference when creating the metric."
            )

        # Handle empty string separately
        if ref == "":
            raise ValueError(
                "Invalid reference string provided. Reference must be a non-empty string."
            )

        value = output if self._case_sensitive else output.lower()
        ref = ref if self._case_sensitive else ref.lower()

        if ref in value:
            return score_result.ScoreResult(value=1.0, name=self.name)

        return score_result.ScoreResult(value=0.0, name=self.name)



import copy
import datetime
import typing

from opik.rest_api import client as rest_client
from opik.rest_api.types.agent_blueprint_public import AgentBlueprintPublic
from opik.api_objects.prompt.text.prompt import Prompt
from opik.api_objects.prompt.chat.chat_prompt import ChatPrompt
from opik.rest_api.types.prompt_version_detail import PromptVersionDetail
from . import type_helpers


def _resolve_prompt_from_commit(
    rest_client_: rest_client.OpikApi, commit: str
) -> typing.Any:
    prompt_detail = rest_client_.prompts.get_prompt_by_commit(commit)
    version_detail = prompt_detail.requested_version
    if version_detail.template_structure == "chat":
        return ChatPrompt.from_fern_prompt_version(
            name=prompt_detail.name, prompt_version=version_detail
        )
    return Prompt.from_fern_prompt_version(
        name=prompt_detail.name, prompt_version=version_detail
    )


def _resolve_prompt_version_from_commit(
    rest_client_: rest_client.OpikApi, commit: str
) -> PromptVersionDetail:
    prompt_detail = rest_client_.prompts.get_prompt_by_commit(commit)
    return prompt_detail.requested_version


def _convert_primitives(
    raw_blueprint: AgentBlueprintPublic,
    field_types: typing.Optional[typing.Dict[str, typing.Any]],
) -> typing.Dict[str, typing.Any]:
    values: typing.Dict[str, typing.Any] = {}
    for param in raw_blueprint.values:
        if field_types and param.key in field_types:
            py_type = field_types[param.key]
        else:
            py_type = type_helpers.backend_type_to_python_type(param.type)

        if py_type is not None:
            values[param.key] = type_helpers.backend_value_to_python_value(
                param.value, param.type, py_type
            )
        else:
            values[param.key] = param.value
    return values


def _is_prompt_field(
    key: str,
    backend_type: str,
    field_types: typing.Optional[typing.Dict[str, typing.Any]],
) -> bool:
    if field_types and key in field_types:
        return type_helpers.is_prompt_type(field_types[key])
    return backend_type == "prompt"


def _is_prompt_version_field(
    key: str,
    backend_type: str,
    field_types: typing.Optional[typing.Dict[str, typing.Any]],
) -> bool:
    if field_types and key in field_types:
        return type_helpers.is_prompt_version_type(field_types[key])
    return backend_type == "prompt_commit"


def _resolve_prompts(
    raw_blueprint: AgentBlueprintPublic,
    values: typing.Dict[str, typing.Any],
    field_types: typing.Optional[typing.Dict[str, typing.Any]],
    rest_client_: rest_client.OpikApi,
) -> None:
    for param in raw_blueprint.values:
        raw_value = values.get(param.key)
        if not isinstance(raw_value, str):
            continue

        if _is_prompt_field(param.key, param.type, field_types):
            values[param.key] = _resolve_prompt_from_commit(rest_client_, raw_value)
        elif _is_prompt_version_field(param.key, param.type, field_types):
            values[param.key] = _resolve_prompt_version_from_commit(
                rest_client_, raw_value
            )


def _resolve_values(
    raw_blueprint: AgentBlueprintPublic,
    field_types: typing.Optional[typing.Dict[str, typing.Any]],
    rest_client_: typing.Optional[rest_client.OpikApi],
) -> typing.Dict[str, typing.Any]:
    values = _convert_primitives(raw_blueprint, field_types)
    if rest_client_:
        _resolve_prompts(raw_blueprint, values, field_types, rest_client_)
    return values


class Blueprint:
    """A specific versioned snapshot of agent config values (read-only)."""

    def __init__(
        self,
        raw_blueprint: AgentBlueprintPublic,
        field_types: typing.Optional[typing.Dict[str, typing.Any]] = None,
        rest_client_: typing.Optional[rest_client.OpikApi] = None,
    ) -> None:
        self._raw = raw_blueprint
        self._values = _resolve_values(raw_blueprint, field_types, rest_client_)

    @property
    def id(self) -> typing.Optional[str]:
        return self._raw.id

    @property
    def name(self) -> typing.Optional[str]:
        return self._raw.name

    @property
    def description(self) -> typing.Optional[str]:
        return self._raw.description

    @property
    def type(self) -> typing.Optional[str]:
        return self._raw.type

    @property
    def envs(self) -> typing.Optional[typing.List[str]]:
        return self._raw.envs

    @property
    def created_by(self) -> typing.Optional[str]:
        return self._raw.created_by

    @property
    def created_at(self) -> typing.Optional[datetime.datetime]:
        return self._raw.created_at

    @property
    def values(self) -> typing.Dict[str, typing.Any]:
        return copy.deepcopy(self._values)

    def get(self, key: str, default: typing.Any = None) -> typing.Any:
        return self._values.get(key, default)

    def __getitem__(self, key: str) -> typing.Any:
        return self._values[key]

    def keys(self) -> typing.KeysView[str]:
        return self._values.keys()



# This file was auto-generated by Fern from our API Definition.

import typing

from ..core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ..core.request_options import RequestOptions
from ..types.alert_page_public import AlertPagePublic
from ..types.alert_public import AlertPublic
from ..types.alert_trigger_write import AlertTriggerWrite
from ..types.alert_write_alert_type import AlertWriteAlertType
from ..types.webhook_examples import WebhookExamples
from ..types.webhook_test_result import WebhookTestResult
from ..types.webhook_write import WebhookWrite
from .raw_client import AsyncRawAlertsClient, RawAlertsClient
from .types.get_webhook_examples_request_alert_type import GetWebhookExamplesRequestAlertType

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class AlertsClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._raw_client = RawAlertsClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> RawAlertsClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        RawAlertsClient
        """
        return self._raw_client

    def find_alerts(
        self,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        sorting: typing.Optional[str] = None,
        filters: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AlertPagePublic:
        """
        Find alerts

        Parameters
        ----------
        page : typing.Optional[int]

        size : typing.Optional[int]

        sorting : typing.Optional[str]

        filters : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AlertPagePublic
            OK

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.alerts.find_alerts()
        """
        _response = self._raw_client.find_alerts(
            page=page, size=size, sorting=sorting, filters=filters, request_options=request_options
        )
        return _response.data

    def create_alert(
        self,
        *,
        webhook: WebhookWrite,
        id: typing.Optional[str] = OMIT,
        name: typing.Optional[str] = OMIT,
        enabled: typing.Optional[bool] = OMIT,
        alert_type: typing.Optional[AlertWriteAlertType] = OMIT,
        metadata: typing.Optional[typing.Dict[str, str]] = OMIT,
        triggers: typing.Optional[typing.Sequence[AlertTriggerWrite]] = OMIT,
        project_id: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Create alert

        Parameters
        ----------
        webhook : WebhookWrite

        id : typing.Optional[str]

        name : typing.Optional[str]

        enabled : typing.Optional[bool]

        alert_type : typing.Optional[AlertWriteAlertType]

        metadata : typing.Optional[typing.Dict[str, str]]

        triggers : typing.Optional[typing.Sequence[AlertTriggerWrite]]

        project_id : typing.Optional[str]
            Optional project scope for this alert. When set, the alert is scoped to the specified project. Do NOT also provide a 'scope:project' trigger config — the system will create it automatically from this field. Sending both project_id and a scope:project trigger config will result in an error.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        from Opik import WebhookWrite
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.alerts.create_alert(webhook=WebhookWrite(url='url', ), )
        """
        _response = self._raw_client.create_alert(
            webhook=webhook,
            id=id,
            name=name,
            enabled=enabled,
            alert_type=alert_type,
            metadata=metadata,
            triggers=triggers,
            project_id=project_id,
            request_options=request_options,
        )
        return _response.data

    def delete_alert_batch(
        self, *, ids: typing.Sequence[str], request_options: typing.Optional[RequestOptions] = None
    ) -> None:
        """
        Delete multiple alerts by their IDs

        Parameters
        ----------
        ids : typing.Sequence[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.alerts.delete_alert_batch(ids=['ids'], )
        """
        _response = self._raw_client.delete_alert_batch(ids=ids, request_options=request_options)
        return _response.data

    def get_alert_by_id(self, id: str, *, request_options: typing.Optional[RequestOptions] = None) -> AlertPublic:
        """
        Get Alert by id

        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AlertPublic
            Alert resource

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.alerts.get_alert_by_id(id='id', )
        """
        _response = self._raw_client.get_alert_by_id(id, request_options=request_options)
        return _response.data

    def update_alert(
        self,
        id_: str,
        *,
        webhook: WebhookWrite,
        id: typing.Optional[str] = OMIT,
        name: typing.Optional[str] = OMIT,
        enabled: typing.Optional[bool] = OMIT,
        alert_type: typing.Optional[AlertWriteAlertType] = OMIT,
        metadata: typing.Optional[typing.Dict[str, str]] = OMIT,
        triggers: typing.Optional[typing.Sequence[AlertTriggerWrite]] = OMIT,
        project_id: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Update alert

        Parameters
        ----------
        id_ : str

        webhook : WebhookWrite

        id : typing.Optional[str]

        name : typing.Optional[str]

        enabled : typing.Optional[bool]

        alert_type : typing.Optional[AlertWriteAlertType]

        metadata : typing.Optional[typing.Dict[str, str]]

        triggers : typing.Optional[typing.Sequence[AlertTriggerWrite]]

        project_id : typing.Optional[str]
            Optional project scope for this alert. When set, the alert is scoped to the specified project. Do NOT also provide a 'scope:project' trigger config — the system will create it automatically from this field. Sending both project_id and a scope:project trigger config will result in an error.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        from Opik import WebhookWrite
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.alerts.update_alert(id_='id', webhook=WebhookWrite(url='url', ), )
        """
        _response = self._raw_client.update_alert(
            id_,
            webhook=webhook,
            id=id,
            name=name,
            enabled=enabled,
            alert_type=alert_type,
            metadata=metadata,
            triggers=triggers,
            project_id=project_id,
            request_options=request_options,
        )
        return _response.data

    def get_webhook_examples(
        self,
        *,
        alert_type: typing.Optional[GetWebhookExamplesRequestAlertType] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> WebhookExamples:
        """
        Get webhook payload examples for all alert event types, optionally filtered by alert type

        Parameters
        ----------
        alert_type : typing.Optional[GetWebhookExamplesRequestAlertType]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        WebhookExamples
            Webhook examples

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.alerts.get_webhook_examples()
        """
        _response = self._raw_client.get_webhook_examples(alert_type=alert_type, request_options=request_options)
        return _response.data

    def test_webhook(
        self,
        *,
        webhook: WebhookWrite,
        id: typing.Optional[str] = OMIT,
        name: typing.Optional[str] = OMIT,
        enabled: typing.Optional[bool] = OMIT,
        alert_type: typing.Optional[AlertWriteAlertType] = OMIT,
        metadata: typing.Optional[typing.Dict[str, str]] = OMIT,
        triggers: typing.Optional[typing.Sequence[AlertTriggerWrite]] = OMIT,
        project_id: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> WebhookTestResult:
        """
        Test alert webhook

        Parameters
        ----------
        webhook : WebhookWrite

        id : typing.Optional[str]

        name : typing.Optional[str]

        enabled : typing.Optional[bool]

        alert_type : typing.Optional[AlertWriteAlertType]

        metadata : typing.Optional[typing.Dict[str, str]]

        triggers : typing.Optional[typing.Sequence[AlertTriggerWrite]]

        project_id : typing.Optional[str]
            Optional project scope for this alert. When set, the alert is scoped to the specified project. Do NOT also provide a 'scope:project' trigger config — the system will create it automatically from this field. Sending both project_id and a scope:project trigger config will result in an error.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        WebhookTestResult
            Webhook test

        Examples
        --------
        from Opik import OpikApi
        from Opik import WebhookWrite
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.alerts.test_webhook(webhook=WebhookWrite(url='url', ), )
        """
        _response = self._raw_client.test_webhook(
            webhook=webhook,
            id=id,
            name=name,
            enabled=enabled,
            alert_type=alert_type,
            metadata=metadata,
            triggers=triggers,
            project_id=project_id,
            request_options=request_options,
        )
        return _response.data


class AsyncAlertsClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._raw_client = AsyncRawAlertsClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> AsyncRawAlertsClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        AsyncRawAlertsClient
        """
        return self._raw_client

    async def find_alerts(
        self,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        sorting: typing.Optional[str] = None,
        filters: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AlertPagePublic:
        """
        Find alerts

        Parameters
        ----------
        page : typing.Optional[int]

        size : typing.Optional[int]

        sorting : typing.Optional[str]

        filters : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AlertPagePublic
            OK

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.alerts.find_alerts()
        asyncio.run(main())
        """
        _response = await self._raw_client.find_alerts(
            page=page, size=size, sorting=sorting, filters=filters, request_options=request_options
        )
        return _response.data

    async def create_alert(
        self,
        *,
        webhook: WebhookWrite,
        id: typing.Optional[str] = OMIT,
        name: typing.Optional[str] = OMIT,
        enabled: typing.Optional[bool] = OMIT,
        alert_type: typing.Optional[AlertWriteAlertType] = OMIT,
        metadata: typing.Optional[typing.Dict[str, str]] = OMIT,
        triggers: typing.Optional[typing.Sequence[AlertTriggerWrite]] = OMIT,
        project_id: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Create alert

        Parameters
        ----------
        webhook : WebhookWrite

        id : typing.Optional[str]

        name : typing.Optional[str]

        enabled : typing.Optional[bool]

        alert_type : typing.Optional[AlertWriteAlertType]

        metadata : typing.Optional[typing.Dict[str, str]]

        triggers : typing.Optional[typing.Sequence[AlertTriggerWrite]]

        project_id : typing.Optional[str]
            Optional project scope for this alert. When set, the alert is scoped to the specified project. Do NOT also provide a 'scope:project' trigger config — the system will create it automatically from this field. Sending both project_id and a scope:project trigger config will result in an error.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        from Opik import WebhookWrite
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.alerts.create_alert(webhook=WebhookWrite(url='url', ), )
        asyncio.run(main())
        """
        _response = await self._raw_client.create_alert(
            webhook=webhook,
            id=id,
            name=name,
            enabled=enabled,
            alert_type=alert_type,
            metadata=metadata,
            triggers=triggers,
            project_id=project_id,
            request_options=request_options,
        )
        return _response.data

    async def delete_alert_batch(
        self, *, ids: typing.Sequence[str], request_options: typing.Optional[RequestOptions] = None
    ) -> None:
        """
        Delete multiple alerts by their IDs

        Parameters
        ----------
        ids : typing.Sequence[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.alerts.delete_alert_batch(ids=['ids'], )
        asyncio.run(main())
        """
        _response = await self._raw_client.delete_alert_batch(ids=ids, request_options=request_options)
        return _response.data

    async def get_alert_by_id(self, id: str, *, request_options: typing.Optional[RequestOptions] = None) -> AlertPublic:
        """
        Get Alert by id

        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AlertPublic
            Alert resource

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.alerts.get_alert_by_id(id='id', )
        asyncio.run(main())
        """
        _response = await self._raw_client.get_alert_by_id(id, request_options=request_options)
        return _response.data

    async def update_alert(
        self,
        id_: str,
        *,
        webhook: WebhookWrite,
        id: typing.Optional[str] = OMIT,
        name: typing.Optional[str] = OMIT,
        enabled: typing.Optional[bool] = OMIT,
        alert_type: typing.Optional[AlertWriteAlertType] = OMIT,
        metadata: typing.Optional[typing.Dict[str, str]] = OMIT,
        triggers: typing.Optional[typing.Sequence[AlertTriggerWrite]] = OMIT,
        project_id: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Update alert

        Parameters
        ----------
        id_ : str

        webhook : WebhookWrite

        id : typing.Optional[str]

        name : typing.Optional[str]

        enabled : typing.Optional[bool]

        alert_type : typing.Optional[AlertWriteAlertType]

        metadata : typing.Optional[typing.Dict[str, str]]

        triggers : typing.Optional[typing.Sequence[AlertTriggerWrite]]

        project_id : typing.Optional[str]
            Optional project scope for this alert. When set, the alert is scoped to the specified project. Do NOT also provide a 'scope:project' trigger config — the system will create it automatically from this field. Sending both project_id and a scope:project trigger config will result in an error.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        from Opik import WebhookWrite
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.alerts.update_alert(id_='id', webhook=WebhookWrite(url='url', ), )
        asyncio.run(main())
        """
        _response = await self._raw_client.update_alert(
            id_,
            webhook=webhook,
            id=id,
            name=name,
            enabled=enabled,
            alert_type=alert_type,
            metadata=metadata,
            triggers=triggers,
            project_id=project_id,
            request_options=request_options,
        )
        return _response.data

    async def get_webhook_examples(
        self,
        *,
        alert_type: typing.Optional[GetWebhookExamplesRequestAlertType] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> WebhookExamples:
        """
        Get webhook payload examples for all alert event types, optionally filtered by alert type

        Parameters
        ----------
        alert_type : typing.Optional[GetWebhookExamplesRequestAlertType]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        WebhookExamples
            Webhook examples

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.alerts.get_webhook_examples()
        asyncio.run(main())
        """
        _response = await self._raw_client.get_webhook_examples(alert_type=alert_type, request_options=request_options)
        return _response.data

    async def test_webhook(
        self,
        *,
        webhook: WebhookWrite,
        id: typing.Optional[str] = OMIT,
        name: typing.Optional[str] = OMIT,
        enabled: typing.Optional[bool] = OMIT,
        alert_type: typing.Optional[AlertWriteAlertType] = OMIT,
        metadata: typing.Optional[typing.Dict[str, str]] = OMIT,
        triggers: typing.Optional[typing.Sequence[AlertTriggerWrite]] = OMIT,
        project_id: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> WebhookTestResult:
        """
        Test alert webhook

        Parameters
        ----------
        webhook : WebhookWrite

        id : typing.Optional[str]

        name : typing.Optional[str]

        enabled : typing.Optional[bool]

        alert_type : typing.Optional[AlertWriteAlertType]

        metadata : typing.Optional[typing.Dict[str, str]]

        triggers : typing.Optional[typing.Sequence[AlertTriggerWrite]]

        project_id : typing.Optional[str]
            Optional project scope for this alert. When set, the alert is scoped to the specified project. Do NOT also provide a 'scope:project' trigger config — the system will create it automatically from this field. Sending both project_id and a scope:project trigger config will result in an error.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        WebhookTestResult
            Webhook test

        Examples
        --------
        from Opik import AsyncOpikApi
        from Opik import WebhookWrite
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.alerts.test_webhook(webhook=WebhookWrite(url='url', ), )
        asyncio.run(main())
        """
        _response = await self._raw_client.test_webhook(
            webhook=webhook,
            id=id,
            name=name,
            enabled=enabled,
            alert_type=alert_type,
            metadata=metadata,
            triggers=triggers,
            project_id=project_id,
            request_options=request_options,
        )
        return _response.data



from typing import Optional, Union

import pydantic


class Attachment(pydantic.BaseModel):
    """
    Represents an Attachment to be added to the Trace or Span.

    Args:
        data: The data to be added to the Attachment. Can be:
            - A file path (str) to an existing file
            - A base64-encoded string (str) representing file content
            - Raw bytes content
        file_name: The custom filename to assign to the data in the attachment.
            If not provided, the original filename of the data will be used.
        content_type: The MIME type of the data to be added to the attachment.
            If not specified, it will be inferred from the data file.
        create_temp_copy: If True, a temporary copy of the file will be created
            before upload. This ensures the file remains available even if the
            original is deleted. The temp file will be deleted after upload.
            Default is True.
    """

    data: Union[str, bytes]
    file_name: Optional[str] = None
    content_type: Optional[str] = None
    create_temp_copy: bool = True



import asyncio
import logging
from typing import Optional, Union, Any, List, Dict

import pydantic

from opik.evaluation.metrics.conversation import helpers as conversation_helpers
from opik.evaluation.metrics.conversation import types as conversation_types
import opik.exceptions as exceptions
from opik.evaluation.metrics import score_result
from opik.evaluation.metrics.conversation.conversation_thread_metric import (
    ConversationThreadMetric,
)
from opik.evaluation.metrics.llm_judges import parsing_helpers
from opik.evaluation.models import base_model, models_factory
from . import schema, templates

LOGGER = logging.getLogger(__name__)


class ConversationalCoherenceMetric(ConversationThreadMetric):
    """
    Calculates the conversational coherence metric for a given conversation thread.
    This metric assesses the coherence and relevance across a series of conversation
    turns by evaluating the consistency in responses, logical flow, and overall
    context maintenance. It evaluates whether the conversation session felt like a
    natural, adaptive, helpful interaction.

    The ``ConversationalCoherenceMetric`` builds a sliding window of dialogue turns for
    each turn in the conversation. It then uses a language model to evaluate whether
    the final `assistant` message within each window is relevant and coherent in
    relation to the preceding conversational context.

    It supports both synchronous and asynchronous operations to
    accommodate the model's operation type. It returns a score between `0.0` and `1.0`,
    where `0.0` indicates a low coherence score and `1.0` indicates a high coherence score.

    Args:
        model: The model to use for
            evaluating the conversation. If a string is provided, it will be used to
            fetch the model from the LiteLLM API. If a base_model.OpikBaseModel is
            provided, it will be used directly. Default is None.
        name: The name of the metric. The default is "conversational_coherence_score".
        include_reason: Whether to include the reason for the score in the
            result. Default is True.
        track: Whether to track the metric. Default is True.
        project_name: The name of the project to track the metric in.
            Default is None.
        window_size: The window size to use for calculating the score. It defines the
            maximal number of historical turns to include in each window when assessing
            the coherence of the current turn in the conversation. Default is 10.
        temperature: The temperature to use for the model. Defaults to 1e-8.

    Example:
        >>> from opik.evaluation.metrics import ConversationalCoherenceMetric
        >>> conversation = [
        >>>     {"role": "user", "content": "Hello!"},
        >>>     {"role": "assistant", "content": "Hi there!"},
        >>>     {"role": "user", "content": "How are you?"},
        >>>     {"role": "assistant", "content": "I'm doing well, thank you!"},
        >>> ]
        >>> metric = ConversationalCoherenceMetric()
        >>> result = metric.score(conversation)
        >>> if result.scoring_failed:
        >>>     print(f"Scoring failed: {result.reason}")
        >>> else:
        >>>     print(result.value)
    """

    def __init__(
        self,
        model: Optional[Union[str, base_model.OpikBaseModel]] = None,
        name: str = "conversational_coherence_score",
        include_reason: bool = True,
        track: bool = True,
        project_name: Optional[str] = None,
        window_size: int = 10,
        temperature: float = 1e-8,
    ):
        super().__init__(
            name=name,
            track=track,
            project_name=project_name,
        )
        self._include_reason = include_reason
        self._window_size = window_size

        self._init_model(model, temperature=temperature)

    def _init_model(
        self, model: Optional[Union[str, base_model.OpikBaseModel]], temperature: float
    ) -> None:
        if isinstance(model, base_model.OpikBaseModel):
            self._model = model
        else:
            self._model = models_factory.get(
                model_name=model, track=self.track, temperature=temperature
            )

    def score(
        self,
        conversation: conversation_types.Conversation,
        **ignored_kwargs: Any,
    ) -> score_result.ScoreResult:
        return self._calculate_score(conversation=conversation)

    async def ascore(
        self,
        conversation: conversation_types.Conversation,
        **ignored_kwargs: Any,
    ) -> score_result.ScoreResult:
        return await self._a_calculate_score(conversation=conversation)

    def _calculate_score(
        self,
        conversation: conversation_types.Conversation,
    ) -> score_result.ScoreResult:
        try:
            turns_windows = (
                conversation_helpers.extract_turns_windows_from_conversation(
                    conversation=conversation, window_size=self._window_size
                )
            )

            verdicts = [
                self._evaluate_conversation(conversation_sliding_window=window)
                for window in turns_windows
            ]

            score = _score_from_verdicts(verdicts=verdicts)
            reason = (
                self._reason_from_verdicts(score=score, verdicts=verdicts)
                if self._include_reason
                else None
            )
            return score_result.ScoreResult(
                name=self.name,
                value=score,
                reason=reason,
            )
        except Exception as e:
            LOGGER.error(f"Failed to calculate conversational coherence score: {e}")
            raise exceptions.MetricComputationError(
                f"Failed to calculate conversational coherence score: {e}"
            ) from e

    async def _a_calculate_score(
        self,
        conversation: conversation_types.Conversation,
    ) -> score_result.ScoreResult:
        try:
            turns_windows = (
                conversation_helpers.extract_turns_windows_from_conversation(
                    conversation=conversation, window_size=self._window_size
                )
            )

            verdicts = await asyncio.gather(
                *[
                    self._a_evaluate_conversation(conversation_sliding_window=window)
                    for window in turns_windows
                ]
            )

            score = _score_from_verdicts(verdicts=verdicts)
            reason = (
                await self._a_reason_from_verdicts(score=score, verdicts=verdicts)
                if self._include_reason
                else None
            )
            return score_result.ScoreResult(
                name=self.name,
                value=score,
                reason=reason,
            )
        except Exception as e:
            LOGGER.error(f"Failed to calculate conversational coherence score: {e}")
            raise exceptions.MetricComputationError(
                f"Failed to calculate conversational coherence score: {e}"
            )

    def _reason_from_verdicts(
        self, score: float, verdicts: List[schema.EvaluateConversationCoherenceResponse]
    ) -> str:
        irrelevancies: List[Dict[str, str]] = _extract_irrelevancies_from_verdicts(
            verdicts
        )
        llm_query = templates.generate_reason(score=score, irrelevancies=irrelevancies)
        model_output = self._model.generate_string(
            input=llm_query, response_format=schema.ScoreReasonResponse
        )
        return _generate_reason_from_model_output(model_output=model_output)

    async def _a_reason_from_verdicts(
        self, score: float, verdicts: List[schema.EvaluateConversationCoherenceResponse]
    ) -> str:
        irrelevancies: List[Dict[str, str]] = _extract_irrelevancies_from_verdicts(
            verdicts
        )
        llm_query = templates.generate_reason(score=score, irrelevancies=irrelevancies)
        model_output = await self._model.agenerate_string(
            input=llm_query, response_format=schema.ScoreReasonResponse
        )
        return _generate_reason_from_model_output(model_output=model_output)

    def _evaluate_conversation(
        self, conversation_sliding_window: conversation_types.Conversation
    ) -> schema.EvaluateConversationCoherenceResponse:
        llm_query = templates.evaluate_conversation(
            sliding_window=conversation_sliding_window
        )

        model_output = self._model.generate_string(
            input=llm_query,
            response_format=schema.EvaluateConversationCoherenceResponse,
        )
        return _evaluate_conversation_from_model_output(model_output=model_output)

    async def _a_evaluate_conversation(
        self, conversation_sliding_window: conversation_types.Conversation
    ) -> schema.EvaluateConversationCoherenceResponse:
        llm_query = templates.evaluate_conversation(
            sliding_window=conversation_sliding_window
        )

        model_output = await self._model.agenerate_string(
            input=llm_query,
            response_format=schema.EvaluateConversationCoherenceResponse,
        )
        return _evaluate_conversation_from_model_output(model_output=model_output)


def _generate_reason_from_model_output(model_output: str) -> str:
    try:
        dict_content = parsing_helpers.extract_json_content_or_raise(model_output)
        return schema.ScoreReasonResponse.model_validate(dict_content).reason
    except pydantic.ValidationError as e:
        LOGGER.warning(
            f"Failed to parse coherence score reason from the LLM output: {model_output}, reason: {e}",
            exc_info=True,
        )
        raise e


def _extract_irrelevancies_from_verdicts(
    verdicts: List[schema.EvaluateConversationCoherenceResponse],
) -> List[Dict[str, str]]:
    return [
        {"message_number": f"{index + 1}", "reason": verdict.reason}
        for index, verdict in enumerate(verdicts)
        if verdict.verdict.strip().lower() == "no" and verdict.reason is not None
    ]


def _evaluate_conversation_from_model_output(
    model_output: str,
) -> schema.EvaluateConversationCoherenceResponse:
    try:
        dict_content = parsing_helpers.extract_json_content_or_raise(model_output)
        return schema.EvaluateConversationCoherenceResponse.model_validate(dict_content)
    except pydantic.ValidationError as e:
        LOGGER.warning(
            f"Failed to parse conversation coherence evaluation results from the LLM output: {model_output}, reason: {e}",
            exc_info=True,
        )
        raise e


def _score_from_verdicts(
    verdicts: List[schema.EvaluateConversationCoherenceResponse],
) -> float:
    if len(verdicts) == 0:
        return 0.0

    relevant_count = sum(v.verdict.strip().lower() != "no" for v in verdicts)
    return relevant_count / len(verdicts)



"""EvaluationSuiteResult and ItemResult types."""

from __future__ import annotations

import dataclasses
from typing import Any, Dict, List, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from opik.evaluation import evaluation_result, test_result
    from opik.evaluation.metrics.score_result import ScoreResult


def is_score_passed(score: ScoreResult) -> bool:
    """Determine whether a score result represents a passing assertion."""
    if score.scoring_failed:
        return False
    return (isinstance(score.value, bool) and score.value) or score.value == 1


@dataclasses.dataclass
class ItemResult:
    """Result for a single evaluation suite item."""

    dataset_item_id: str
    """The ID of the dataset item."""

    passed: bool
    """Whether this item passed based on its execution policy."""

    runs_passed: int
    """Number of runs that passed for this item."""

    runs_total: int
    """Total number of runs completed for this item."""

    configured_runs_per_item: int
    """Configured runs_per_item from the execution policy."""

    pass_threshold: int
    """Minimum passing runs required (from execution policy)."""

    test_results: List[test_result.TestResult]
    """Individual test results for each run of this item."""


class EvaluationSuiteResult:
    """
    Result of running an evaluation suite.

    Contains pass/fail status for each item based on execution policy,
    as well as overall suite pass/fail status.
    """

    def __init__(
        self,
        items_passed: int,
        items_total: int,
        item_results: Dict[str, ItemResult],
        evaluation_result_: evaluation_result.EvaluationResult,
        suite_name: Optional[str] = None,
        total_time: Optional[float] = None,
    ) -> None:
        self._items_passed = items_passed
        self._items_total = items_total
        self._item_results = item_results
        self._evaluation_result = evaluation_result_
        self._suite_name = suite_name
        self._total_time = total_time

    @property
    def all_items_passed(self) -> bool:
        """Whether all items in the suite passed."""
        return self._items_passed == self._items_total

    @property
    def items_passed(self) -> int:
        """Number of items that passed."""
        return self._items_passed

    @property
    def items_total(self) -> int:
        """Total number of items evaluated."""
        return self._items_total

    @property
    def item_results(self) -> Dict[str, ItemResult]:
        """Results for each item, keyed by dataset_item_id."""
        return self._item_results

    @property
    def pass_rate(self) -> Optional[float]:
        """Pass rate: items_passed / items_total (0.0 to 1.0), or None if no items."""
        if self._items_total == 0:
            return None
        return self._items_passed / self._items_total

    @property
    def suite_name(self) -> Optional[str]:
        """The name of the evaluation suite."""
        return self._suite_name

    @property
    def total_time(self) -> Optional[float]:
        """Total evaluation time in seconds."""
        return self._total_time

    @property
    def experiment_id(self) -> str:
        """The experiment ID."""
        return self._evaluation_result.experiment_id

    @property
    def experiment_name(self) -> Optional[str]:
        """The experiment name."""
        return self._evaluation_result.experiment_name

    @property
    def experiment_url(self) -> Optional[str]:
        """URL to view the experiment."""
        return self._evaluation_result.experiment_url

    def to_report_dict(self) -> Dict[str, Any]:
        """Convert the result to a structured report dictionary."""
        from datetime import datetime, timezone

        items: List[Dict[str, Any]] = []

        for item_id, item_result in self._item_results.items():
            runs: List[Dict[str, Any]] = []

            for test_result_ in item_result.test_results:
                assertions: List[Dict[str, Any]] = []
                for score in test_result_.score_results:
                    assertion: Dict[str, Any] = {
                        "name": score.name,
                        "passed": is_score_passed(score),
                        "value": score.value,
                        "scoring_failed": score.scoring_failed,
                    }
                    if score.reason is not None:
                        assertion["reason"] = score.reason
                    if score.metadata is not None:
                        assertion["metadata"] = score.metadata
                    assertions.append(assertion)

                run_passed = (
                    all(a["passed"] for a in assertions) if assertions else True
                )

                run: Dict[str, Any] = {
                    "trial_id": test_result_.trial_id,
                    "passed": run_passed,
                    "input": test_result_.test_case.task_output.get("input"),
                    "output": test_result_.test_case.task_output.get("output"),
                    "assertions": assertions,
                }
                if test_result_.test_case.trace_id:
                    run["trace_id"] = test_result_.test_case.trace_id
                if test_result_.task_execution_time is not None:
                    run["task_execution_time_seconds"] = round(
                        test_result_.task_execution_time, 3
                    )
                if test_result_.scoring_time is not None:
                    run["scoring_time_seconds"] = round(test_result_.scoring_time, 3)
                runs.append(run)

            items.append(
                {
                    "dataset_item_id": item_id,
                    "passed": item_result.passed,
                    "runs_passed": item_result.runs_passed,
                    "execution_policy": {
                        "runs_per_item": item_result.configured_runs_per_item,
                        "pass_threshold": item_result.pass_threshold,
                    },
                    "runs": runs,
                }
            )

        report: Dict[str, Any] = {
            "suite_passed": self.all_items_passed,
            "items_passed": self._items_passed,
            "items_total": self._items_total,
            "pass_rate": self.pass_rate,
            "experiment_id": self.experiment_id,
        }

        if self._suite_name is not None:
            report["suite_name"] = self._suite_name

        if self.experiment_name is not None:
            report["experiment_name"] = self.experiment_name

        if self.experiment_url is not None:
            report["experiment_url"] = self.experiment_url

        if self._total_time is not None:
            report["total_time_seconds"] = round(self._total_time, 3)

        report["generated_at"] = datetime.now(timezone.utc).isoformat()
        report["items"] = items

        return report

    def to_dict(self) -> Dict[str, Any]:
        """Alias for to_report_dict()."""
        return self.to_report_dict()



from typing import Any, Callable, Dict, List, Union

from . import test_result
from .metrics import score_result

LLMTask = Callable[[Dict[str, Any]], Dict[str, Any]]

ScoringKeyMappingType = Dict[str, Union[str, Callable[[Dict[str, Any]], Any]]]

ExperimentScoreFunction = Callable[
    [List[test_result.TestResult]],
    Union[score_result.ScoreResult, List[score_result.ScoreResult]],
]



from typing import Dict, Any, Optional, TYPE_CHECKING
import dataclasses

if TYPE_CHECKING:
    from opik.api_objects.dataset import dataset_item


@dataclasses.dataclass
class TestCase:
    trace_id: str
    dataset_item_id: str
    task_output: Dict[str, Any]
    dataset_item_content: Dict[str, Any] = dataclasses.field(default_factory=dict)
    mapped_scoring_inputs: Optional[Dict[str, Any]] = None
    dataset_item: Optional["dataset_item.DatasetItem"] = None



"""Execution policy type and default for dataset evaluation."""

from typing import TypedDict


class ExecutionPolicy(TypedDict, total=False):
    """
    Execution policy for evaluation suite items.

    Attributes:
        runs_per_item: Number of times to run evaluation per item.
        pass_threshold: Minimum number of passing runs required for item to pass.
    """

    runs_per_item: int
    pass_threshold: int


DEFAULT_EXECUTION_POLICY: ExecutionPolicy = {
    "runs_per_item": 1,
    "pass_threshold": 1,
}
